"""
Keeps all the space names for code consistency. Makes refactoring much easier
"""
REFLECTANCE_SPECTRUM = 'Spectrum'
XYZ = 'CIEXYZ'
XYY = 'xyY'
NORMALIZED_XYZ = 'XYZ_r'
LAB = 'CIELAB'
LCH = 'CIELCH'
LINEAR_RGB = 'lRGB'
RGB = 'RGB'
HSL = 'HSL'
HSI = 'HSI'
HCY = 'HCY'
HSV = 'HSV'



